
"""Module for holding functions used to evovle the rotation of stars."""

# Imports for standard stuff needed here
import numpy as np
import math
import copy

# Imports for mors modules
import mors.miscellaneous as misc
import mors.physicalmodel as phys
import mors.parameters as params
import mors.stellarevo as SE
import sys

AgeMinDefault = 1.0       # when to start evolution (Myr)
AgeMaxDefault = 5000.0    # when to end evolution (Myr)

def FitRotation(Mstar=None,Age=None,Omega=None,AgeMin=None,params=params.paramsDefault,StarEvo=None):
    """Takes stellar mass, age, and surface rotation rate, returns initial rotation rate of star. 
    
    This function can be used if the age and surface rotation rate of the star is known and the user wants to get
    the corresponding initial (usually 1 Myr) rotation rate for the star. The star's mass, age, and rotation rates 
    must be input. The starting age to get the initial rotation rate for can be set, and by default this is set to
    AgeMinFit in the parameters and it is not recommended to change this. In addition, model parameters and an instance
    of the StarEvo class can be input. The return value of this function should be the initial rotation rate but if 
    the input Omega is below the minimum that can be fit, it will be -1, if it is above the maximum that can be 
    fit, it will be -2, and if it is in the range but no solution could be found it will be -3.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of the star in Myr. 
    Omega : float 
        Rotation rate in OmegaSun (=2.67e-6 rad/s).
    AgeMin : float , optional
        Starting age (default = 1 Myr).
    params : dict , optional
        Parameters to determine behavior of code (default given in parameters.py).
    StarEvo : mors.stellarevo.StarEvo
        Instance of StarEvo class holding stellar evolution models to use.
    
    Returns
    ----------
    Omega0 : dict
        Starting rotation rate in OmegaSun (=2.67e-6 rad/s) or negative if cannot get initial rotation.
    
    """
    
    #---------------------------------------------------------------------
        
    # Make sure stellar mass is set
    if Mstar is None: 
        misc._PrintErrorKill("argument Mstar must be set in call to function")
    
    # Make sure initial rotation is set correctly
    if Omega is None: 
        misc._PrintErrorKill("argument Omega must be set in call to function")
    
    # Make sure AgeMin is set
    if AgeMin is None:
        AgeMin = params['AgeMinFit']
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Make sure this exact stellar mass is loaded (will do nothing if it already is)
    StarEvo.LoadTrack(Mstar)
        
    #---------------------------------------------------------------------
    
    # First thing to be checked is if the model can actually fit this rotation rate at this age since
    # we have defined minimum and maximum initial rotation rates in the parameters, given by
    
    # Set starting Omega0Min and Omega0Max for fitting
    Omega0Min = params['Omega0FitMin']
    Omega0Max = params['Omega0FitMax']
    
    # Get minimum rotation rate at Age corresponding to Omega0Min
    Tracks = EvolveRotation(Mstar=Mstar,Omega0=Omega0Min,AgeMin=params['AgeMinFit'],AgeMax=Age,params=params,StarEvo=StarEvo)
    OmegaMin = Tracks['OmegaEnv'][-1]
    
    # Get maximum rotation rate at Age corresponding to Omega0Max
    Tracks = EvolveRotation(Mstar=Mstar,Omega0=Omega0Max,AgeMin=params['AgeMinFit'],AgeMax=Age,params=params,StarEvo=StarEvo)
    OmegaMax = Tracks['OmegaEnv'][-1]
    
    # Check that Omega is between OmegaMin and OmegaMax
    if ( Omega < OmegaMin ):
        return -1
    if ( Omega > OmegaMax ):
        return -2
    
    #---------------------------------------------------------------------
    
    # Initially assume not found solution
    found = False
        
    # Start looping until solution has been found
    for iStep in range(0,params['nStepMaxFit']):
            
        # Get middle Omega0
        Omega0Mid = 0.5 * ( Omega0Min + Omega0Max )
        
        # Get rotation at mid spot
        Tracks = EvolveRotation(Mstar=Mstar,Omega0=Omega0Mid,AgeMin=params['AgeMinFit'],AgeMax=Age,params=params,StarEvo=StarEvo)
        OmegaMid = Tracks['OmegaEnv'][-1]
        
        # Check if OmegaMid is close enough to the desired answer
        if ( abs(OmegaMid/Omega-1.0) < params['toleranceFit'] ):
            found = True
            Omega0 = Omega0Mid
            break
        
        # Check if Omega0Min and Omega0Max are is close enough together (closest they will get)
        if ( abs(Omega0Min/Omega0Max-1.0) < params['toleranceFit'] ):
            found = True
            Omega0 = Omega0Mid
            break
    
        # Change either min or max Omega0
        # Note that we already know that the solution is between the min and max
        if ( Omega < OmegaMid ):
            Omega0Max = Omega0Mid
            OmegaMax = OmegaMid
        else:
            Omega0Min = Omega0Mid
            OmegaMin = OmegaMid
    
    # Return -3 if not found result
    if not found:
        return -3
  
    #---------------------------------------------------------------------
    
    return Omega0

def EvolveRotation(Mstar=None,Omega0=None,OmegaEnv0=None,OmegaCore0=None,AgeMin=None,AgeMax=None,AgesOut=None,params=params.paramsDefault,StarEvo=None):
    """Takes stellar mass and rotation rate, evolves rotation. 
    
    This is the main function for rotational evolution calculations. It will take some basic stellar parameters and
    evolve rotation between two ages.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Omega0 : float , optional
        Starting rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaEnv0 : float , optional
        Starting envelope rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaCore0 : float , optional
        Starting envelope rotation rate in OmegaSun (=2.67e-6 rad/s).
    AgeMin : float , optional
        Starting age (default = 1 Myr).
    AgeMax : float , optional
        Final age (default = 5 Gyr).
    AgesOut : float or numpy.ndarray , optional
        Ages to output data for in Myr.
    params : dict , optional
        Parameters to determine behavior of code (default given in parameters.py).
    StarEvo : mors.stellarevo.StarEvo
        Instance of StarEvo class holding stellar evolution models to use.
    
    Returns
    ----------
    Tracks : dict
        Dictionary with evolutionary tracks and related information. 
    
    """
    
    # Make sure stellar mass is set
    if Mstar is None: 
        misc._PrintErrorKill("argument Mstar must be set in call to function")
    
    # Make sure initial rotation is set correctly
    if Omega0 is None: 
        if ( ( OmegaEnv0 is None ) or ( OmegaCore0 is None ) ):
            misc._PrintErrorKill("if Omega0 is not set then both OmegaEnv0 and OmegaCore0 must be set in call to function")
    else:
        if not ( ( OmegaEnv0 is None ) and ( OmegaCore0 is None ) ):
            misc._PrintErrorKill("if Omega0 is set then neither OmegaEnv0 nor OmegaCore0 can be set in call to function")
        
    # Make sure AgeMin is set
    if AgeMin is None:
        AgeMin = params['AgeMinDefault']
        
    # Make sure AgeMax is set
    if AgeMax is None:
        AgeMax = params['AgeMaxDefault']
        
    # Make editable and np array form of AgesOut and use this from now on
    if not AgesOut is None:
        if isinstance(AgesOut,(float,int)):
            AgesOut2 = np.array([AgesOut])
        else:
            AgesOut2 = copy.deepcopy(AgesOut)
    else:
        AgesOut2 = None
    
    # If AgesOut has been set, use the final age from that as ending age
    if not AgesOut2 is None:
        AgeMax = AgesOut2[-1]
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Make sure this exact stellar mass is loaded (will do nothing if it already is)
    StarEvo.LoadTrack(Mstar)
    
    # Starting values for time
    iAge = 0
    Age = AgeMin
    
    # Starting timestep in Myr
    dAge = 0.5
    
    # If AgesOut was set, make sure no ages are below AgeMin
    if not AgesOut2 is None:
        AgesOut2 = np.delete( AgesOut2 , np.where(AgesOut2<AgeMin) )
        
    # Starting rotation rates
    if Omega0 is None:
        OmegaEnv = OmegaEnv0
        OmegaCore = OmegaCore0
    else:
        OmegaEnv = Omega0
        OmegaCore = Omega0
        
    # Make dictionary for holding tracks and add initial values
    Tracks = _CreateTracks(Mstar,Age,dAge,OmegaEnv,OmegaCore,params,StarEvo)
        
    # Start looping, end only when end time is reached (additional 0.999 factor to stop code getting stuck)
    while ( Age < 0.999*AgeMax ):
        
        # Make sure timestep is not too long
        dAge , dAgeMax = _dAgeCalc(dAge,Age,AgeMax,AgesOut2,params)
        
        # Do timestep
        dAge , dAgeNew , OmegaEnv , OmegaCore = EvolveRotationStep( Mstar=Mstar , Age=Age , 
                                                                OmegaEnv=OmegaEnv , OmegaCore=OmegaCore , 
                                                                dAgeMax=dAgeMax , dAge=dAge , 
                                                                params=params , StarEvo=StarEvo )
        
        # Update time
        iAge += 1
        Age += dAge
        
        # Work out if should add this age to output
        AgesOut2 , shouldAppend = _shouldAppend(Age,AgesOut2)
            
        # Add current state to tracks if should output
        if shouldAppend:
            Tracks = _AppendTracks(Tracks,Mstar,Age,dAge,OmegaEnv,OmegaCore,params,StarEvo)
        
        # Check for bad data
        _CheckBadData(Tracks)
        
        # Get new age step
        dAge = dAgeNew
        
        # Determine if too many steps taken
        if ( iAge >= params['nStepMax'] ):
            misc._PrintErrorKill("too many timesteps taken")
        
    # In Tracks, replace Mstar array with simple Mstar float
    Tracks['Mstar'] = Mstar
    Tracks['nAge'] = len(Tracks['Age'])
    
    return Tracks

def _dAgeCalc(dAge,Age,AgeMax,AgesOut,params):
    """Takes information about age in evolutionary calculation, returns age step to use."""
    
    # Get maximum timestep in Myr
    dAgeMax = AgeMax - Age
    
    # If AgesOut has been set, work out maximum from that
    if not AgesOut is None:
            
        # If AgesOut is just a number (float or int) then it is easy
        if isinstance(AgesOut,(float,int)):
            
            # Get new maximum age
            dAgeMax = AgesOut - Age
            
        else:
            
            # Get index of next age in AgesOut after current Age
            index = misc._getIndexGT(AgesOut,Age)
            
            # Get new maximum age
            dAgeMax = AgesOut[index] - Age
                
    # Make sure timestep is not too long
    dAge = min( dAge , dAgeMax )
    
    # Make sure dAge is within range in parameter file
    dAge = max( dAge , params['dAgeMin'] )
    dAge = min( dAge , params['dAgeMax'] )
    
    return dAge , dAgeMax

def _shouldAppend(Age,AgesOut):
    """Takes information about age in evolutionary calculation, returns if age should be output."""
    
    # If AgesOut is not set, then should always output
    if AgesOut is None:
        return AgesOut , True
    
    # If AgesOut is set, work out if this is one of the output ages
    if not AgesOut is None:
            
        # If AgesOut is just a number (float or int) then it is easy
        if isinstance(AgesOut,(float,int)):
            
            # Check if close enough to AgesOut to output
            if ( abs(Age/float(AgesOut))-1.0 < 1.0e-6 ):
                return AgesOut , True
            else:
                return AgesOut , False
            
        else:
            
            # Get smallest distance between age in AgesOut and current Age
            indexMin = np.argmin( np.abs(AgesOut-Age) )
            deltaAgeMin = np.min( np.abs(AgesOut-Age) )
            
            # Since we are removing this output age each time, indexMin should always be zero
            if not indexMin == 0:
                misc._PrintErrorKill("indexMin is not zero, probably because the input AgesOut is not in ascending order")
            
            # If age is close enough to element in AgesOut then remove this element and return True
            if deltaAgeMin < 1.0e-3:
                AgesOut = np.delete(AgesOut,indexMin)
                return AgesOut , True
            
            else:
                return AgesOut , False
  
    return AgesOut , False

def EvolveRotationStep(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,dAgeMax=None,dAge=None,params=params.paramsDefault,StarEvo=None):
    """Evolves rotation by a single step.
    
    This is the main function to be used to do timestep updates of the core and envelope rotation rates
    for a star given its basic parameters. The five arguments Mstar, Age, OmegaEnv, OmegaCore, 
    and either dAgeMax or dAge must be set in the call to this function. For certain solvers, the timestep 
    is determined automatically and this value is returned along with the updated envelope and
    core rotation rates
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float , optional
        Age of star.
    OmegaEnv : float 
        Envelope rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaCore : float 
        Core rotation rate in OmegaSun (=2.67e-6 rad/s).
    dAgeMax : float , optional
        Maximum timestep to take in Myr.
    dAgeMax : float , optional
        Timestep to take in Myr.
    
    Returns
    ----------
    dAge : numpy.ndarray
        Age step in Myr.
    OmegaEnv : numpy.ndarray
        Envelope rotation rates in OmegaSun (=2.67e-6 rad/s).
    OmegaCore : numpy.ndarray
        Core rotation rates in OmegaSun (=2.67e-6 rad/s).
    
    """
    
    # Make sure parameters are set
    if Mstar is None: 
        misc._PrintErrorKill("argument Mstar must be set in call to function")
    if Age is None: 
        misc._PrintErrorKill("argument Age must be set in call to function")
    if OmegaEnv is None: 
        misc._PrintErrorKill("argument OmegaEnv must be set in call to function")
    if OmegaCore is None: 
        misc._PrintErrorKill("argument OmegaCore must be set in call to function")
    if ( ( dAgeMax is None ) and ( dAge is None ) ): 
        misc._PrintErrorKill("either dAge or dAgeMax argument must be set in call to function")
        
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
        
    # Do step based on method used
    if ( params['TimeIntegrationMethod'] == 'ForwardEuler' ):
        
        dAge , dAgeNew , OmegaEnv , OmegaCore =  _EvolveRotationStepFE( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , 
                                                                    OmegaCore=OmegaCore , dAge=dAge , params=params , StarEvo=StarEvo )
    
    elif ( params['TimeIntegrationMethod'] == 'RungeKutta4' ):
        
        dAge , dAgeNew , OmegaEnv , OmegaCore =  _EvolveRotationStepRK4( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , 
                                                                        OmegaCore=OmegaCore , dAge=dAge , params=params , StarEvo=StarEvo )
    
    elif ( params['TimeIntegrationMethod'] == 'RungeKuttaFehlberg' ):
        
        dAge , dAgeNew , OmegaEnv , OmegaCore =  _EvolveRotationStepRKF( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , 
                                                                        OmegaCore=OmegaCore , dAge=dAge , dAgeMax=dAgeMax ,
                                                                        params=params , StarEvo=StarEvo )
        
    elif ( params['TimeIntegrationMethod'] == 'Rosenbrock' ):
        
        dAge , dAgeNew , OmegaEnv , OmegaCore =  _EvolveRotationStepRB( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , 
                                                                        OmegaCore=OmegaCore , dAge=dAge , dAgeMax=dAgeMax ,
                                                                        params=params , StarEvo=StarEvo )
    
    else:
        misc._PrintErrorKill("invalid value of TimeIntegrationMethod in parameters")
    
    return ( dAge , dAgeNew , OmegaEnv , OmegaCore )

def _EvolveRotationStepFE(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,dAge=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, evolves by timestep using forward Euler method."""
    
    # Get rates of change
    dOmegaEnvdt , dOmegaCoredt = phys.dOmegadt( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , OmegaCore=OmegaCore , params=params , StarEvo=StarEvo )
    
    # Do update
    OmegaEnv += dAge * dOmegaEnvdt
    OmegaCore += dAge * dOmegaCoredt
    
    return ( dAge , dAge , OmegaEnv , OmegaCore )

def _EvolveRotationStepRK4(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,dAge=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, evolves by timestep using classical Runge-Kutta method."""
    
    # Get rates of change
    k1Env , k1Core = phys.dOmegadt( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , OmegaCore=OmegaCore , params=params , StarEvo=StarEvo )
    k2Env , k2Core = phys.dOmegadt( Mstar=Mstar , Age=Age+0.5*dAge , OmegaEnv=OmegaEnv+0.5*dAge*k1Env , OmegaCore=OmegaCore+0.5*dAge*k1Core , params=params , StarEvo=StarEvo )
    k3Env , k3Core = phys.dOmegadt( Mstar=Mstar , Age=Age+0.5*dAge , OmegaEnv=OmegaEnv+0.5*dAge*k2Env , OmegaCore=OmegaCore+0.5*dAge*k2Core , params=params , StarEvo=StarEvo )
    k4Env , k4Core = phys.dOmegadt( Mstar=Mstar , Age=Age+dAge , OmegaEnv=OmegaEnv+dAge*k3Env , OmegaCore=OmegaCore+dAge*k3Core , params=params , StarEvo=StarEvo )
        
    # Do update
    OmegaEnv += dAge * ( k1Env + 2.0*k2Env + 2.0*k3Env + k4Env ) / 6.0
    OmegaCore += dAge * ( k1Core + 2.0*k2Core + 2.0*k3Core + k4Core ) / 6.0
    
    return ( dAge , dAge , OmegaEnv , OmegaCore )

def _EvolveRotationStepRKF(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,dAgeMax=None,dAge=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, evolves by timestep using Runge-Kutta-Fehlberg method."""
    
    # Start with dAge as timestep
    dAgeNew = dAge
    
    # Setup array to hold integration values
    X = np.array([OmegaEnv,OmegaCore])
        
    # Start iterating until got accuracy desired
    while True:
        
        # Set dAge
        dAge = dAgeNew
        
        # k1
        k1 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age , OmegaEnv=X[0] , OmegaCore=X[1] , params=params , StarEvo=StarEvo ) )
        
        # k2
        Age2 = Age + (1.0/5.0)*dAge
        X2 = X + (1.0/5.0)*dAge*k1
        k2 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age2 , OmegaEnv=X2[0] , OmegaCore=X2[1] , params=params , StarEvo=StarEvo ) )
        
        # k3
        Age3 = Age + (3.0/10.0)*dAge
        X3 = X + (3.0/40.0)*dAge*k1 + (9.0/40.0)*dAge*k2
        k3 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age3, OmegaEnv=X3[0] , OmegaCore=X3[1] , params=params , StarEvo=StarEvo ) )
        
        # k4
        Age4 = Age + (3.0/5.0)*dAge
        X4 = X + (3.0/10.0)*dAge*k1 - (9.0/10.0)*dAge*k2 + (6.0/5.0)*dAge*k3
        k4 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age4, OmegaEnv=X4[0] , OmegaCore=X4[1] , params=params , StarEvo=StarEvo ) )
        
        # k5
        Age5 = Age + dAge
        X5 = X - (11.0/54.0)*dAge*k1 + (5.0/2.0)*dAge*k2 - (70.0/27.0)*dAge*k3 + (35.0/27.0)*dAge*k4
        k5 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age5, OmegaEnv=X5[0] , OmegaCore=X5[1] , params=params , StarEvo=StarEvo ) )
        
        # k6
        Age6 = Age + (7.0/8.0)*dAge
        X6 = X + (1631.0/55296.0)*dAge*k1 + (175.0/512.0)*dAge*k2 + (575.0/13824.0)*dAge*k3 + (44275.0/110592.0)*dAge*k4 + (253.0/4096.0)*dAge*k5
        k6 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age6, OmegaEnv=X6[0] , OmegaCore=X6[1] , params=params , StarEvo=StarEvo ) )
        
        # Calculate candidate update
        Xnew = X + (37.0/378.0)*dAge*k1 + (250.0/621.0)*dAge*k3 + (125.0/594.0)*dAge*k4 + (512.0/1771.0)*dAge*k6 
        
        # Calculate forth order update
        Xforth = X + (2825.0/27648.0)*dAge*k1 + (18575.0/48384.0)*dAge*k3 + (13525.0/55296.0)*dAge*k4 + (277.0/14336.0)*dAge*k5 + (1.0/4.0)*dAge*k6
        
        # Get error in this timestep for each species and grid point
        Delta =  Xnew - Xforth
                
        # Get the factor by which to change dt
        # Need to take care here if Delta=0 since this can happen if dX/dt=0 over the entire timestep
        if ( np.min(np.abs(Delta)) == 0.0 ):
            dAgeFactor = 1.5
        else:
            dAgeFactor = np.min( np.abs( params['DeltaDesired']*X / Delta ) )
            
        # Get new dt
        dAgeNew = 0.9 * dAge * dAgeFactor**0.2
        
        # Make sure dt is not too big
        dAgeNew = min( dAgeNew , dAgeMax )
        
        # Check for stopping of iteration
        if ( dAgeFactor > 1.0 ):
            break
        
    # Save new estimate
    OmegaEnv = Xnew[0]
    OmegaCore = Xnew[1]
    
    return dAge , dAgeNew , OmegaEnv , OmegaCore

def _EvolveRotationStepRB(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,dAgeMax=None,dAge=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, evolves by timestep using Runge-Kutta-Fehlberg method."""
    
    # Get coefficients for solver
    CoefficientsRB = params['CoefficientsRB']
    
    # Start with dAge as timestep
    dAgeNew = dAge
    
    # Setup array to hold integration values
    X = np.array([OmegaEnv,OmegaCore])
    nVar = len(X)
    
    # Get Jacobian and assume it is a constant over timestep
    Jac = _JacobianRB( Mstar , Age+dAge , X , nVar , params , StarEvo )
  
    # Start iterating until got accuracy desired
    while True:
        
        # Set dAge
        dAge = dAgeNew
        
        # Calculate the k coefficients
        kCoeff = _kCoeffRB( Mstar , Age , dAge , X , Jac , nVar , CoefficientsRB , params , StarEvo )
        
        # Get updated values
        Xnew = copy.deepcopy(X)
        for i in range(0,CoefficientsRB['s']):
            Xnew += CoefficientsRB['b'][i] * kCoeff[:,i]
    
        # Get less accurate updated values
        Xnew2 = copy.deepcopy(X)
        for i in range(0,CoefficientsRB['s']):
            Xnew2 += CoefficientsRB['b2'][i] * kCoeff[:,i]
        
        # Get error in this timestep for each species and grid point
        Error = np.sqrt( np.sum( ( ( Xnew - Xnew2 ) / params['DeltaDesired'] )**2.0 ) / float(nVar) ) 
        
        # Get the factor by which to change dt (if Error is zero then just increase dAge by 50%)
        if not ( Error == 0.0 ):
            dAgeFactor = min( 10.0 , max( 0.1 , 0.9/Error**(1.0/CoefficientsRB['order']) ) )
        else:
            dAgeFactor = 1.5
        
        # get new dt
        dAgeNew = 0.99 * dAge * dAgeFactor
        
        # Make sure dt is not too big
        dAgeNew = min( dAgeNew , dAgeMax )
        
        # Check for stopping of iteration
        if ( dAgeFactor > 1.0 ):
            break
            
    # Save new estimate
    OmegaEnv = Xnew[0]
    OmegaCore = Xnew[1]
    
    return dAge , dAgeNew , OmegaEnv , OmegaCore

def _JacobianRB(Mstar,Age,X,nVar,params,StarEvo):
    """Calculates Jacobian d(dOmega/dt)/dOmega terms for Rosenbrock solver."""
    
    # Make array
    Jac = np.zeros((nVar,nVar))
    
    # Loop over elements and fill in Jacobian
    for iVar in range(0,nVar):
        
        # Set perturbed X values to X
        X1 = copy.deepcopy(X)
        X2 = copy.deepcopy(X)
        
        # Perturb this variable
        X1[iVar] = X1[iVar] - params['deltaJac'] * X1[iVar]
        X2[iVar] = X2[iVar] + params['deltaJac'] * X2[iVar]
        
        # Get rates based on pertubed quantities
        dXdt1 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age, OmegaEnv=X1[0] , OmegaCore=X1[1] , params=params , StarEvo=StarEvo ) )
        dXdt2 = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age, OmegaEnv=X2[0] , OmegaCore=X2[1] , params=params , StarEvo=StarEvo ) )
        
        # Fill in this column of Jacobian
        Jac[:,iVar] = ( dXdt2[:] - dXdt1[:] ) / ( X2[iVar] - X1[iVar] )
        
    return Jac

def _kCoeffRB(Mstar,Age,dAge,X,Jac,nVar,CoefficientsRB,params,StarEvo):
    """Calculates kCoeff for Rosenbrock solver."""
    
    # Make array for holding result
    kCoeff = np.zeros((nVar,CoefficientsRB['s']))
    
    # Get the function involving the Jacobian, given by (I - dt * gamma_ii * J)
    # note: since all gamma_ii coefficients are equal, this function only needs to
    #       be calculated once and can then be used for all values of i
    # however, it must be recalculated if the timestep is changed
    
    # Set the function to -dt*J for all elements
    JacFunc = - dAge * CoefficientsRB['gamma'][0,0] * Jac
    
    # Loop over diagonal elements and add 1 to each
    for iVar in range(0,nVar):
        JacFunc[iVar,iVar] += 1.0
        
    # Get k1
    dXdt = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age+dAge, OmegaEnv=X[0] , OmegaCore=X[1] , params=params , StarEvo=StarEvo ) )
    RateFunction = dAge * dXdt
    kCoeff[:,0] = _GaussianElimination( JacFunc , RateFunction )
        
    # Get k2,...ks
    for i in range(1,CoefficientsRB['s']):
        
        # Get intermediate values
        Xmid = copy.deepcopy(X)
        for i2 in range(0,i):
            Xmid += CoefficientsRB['alpha'][i,i2] * kCoeff[:,i2]
            
        # Get intermediate rates
        dXdt = np.array( phys.dOmegadt( Mstar=Mstar , Age=Age+dAge, OmegaEnv=Xmid[0] , OmegaCore=Xmid[1] , params=params , StarEvo=StarEvo ) )
            
        # Get the rate function
        RateFunction = dAge * dXdt
        for i2 in range(0,i):
            for iVar in range(0,nVar):
                RateFunction[iVar] += dAge * CoefficientsRB['gamma'][i,i2] * np.sum( Jac[iVar,:] * kCoeff[:,i2] )
            
        # Get k value for this step
        kCoeff[:,i] = _GaussianElimination( JacFunc , RateFunction )
  
    return kCoeff

def _GaussianElimination(Ain,B):
    """Solves the equation A*X = B using Gaussian elimination."""
    
    # This is meant to be used when getting the k coefficients in the Rosenbrock
    # method. This subroutine is able to be relatively simple because the NxN 
    # matrix A is given by I-dt*gamma*J, where I is the identity matrix. It is 
    # highly unlikely that any of the diagonal terms in A are zero, since that
    # would only really be the case if a diagonal term in J is exactly 1.0/(dt*gamma),
    # which would be quite a coincidence. If any diagonal terms in A are zero,
    # this subroutine will give a nonsense result. I have added an if statement
    # below to catch these errors, so the user will at least know what is going on.
    
    #---------------------------------------
    # GET N AND ALLOCATE X AND A
    
    N = len(B)
    X = np.zeros(N)
    A = np.zeros((N,N))
    
    # ---------------------------------------
    # COPY A and B
    
    A[:,:] = Ain[:,:]
    X[:] = B[:]
    
    #---------------------------------------
    # GET IN ROW ECHELON FORMAT
        
    # Loop over rows
    for i in range(0,N):
    
        # Make sure diagonal term not zero
        if ( A[i,i] == 0.0 ):
            misc._PrintErrorKill("diagonal term in zero")
        
        # Divide row by value in diagonal element
        X[i] = X[i] / A[i,i]
        A[i,i:N] = A[i,i:N] / A[i,i]
        
        # Get rid of zeros in columns below this diagonal
        # unless doing final row
        if ( i < N-1 ):
            for k in range(i+1,N):
                if ( A[k,i] != 0.0 ):
                    X[k] += - X[i] * A[k,i]
                    A[k,:] += - A[i,:] * A[k,i]
        
    #---------------------------------------
    # GET IN REDUCED ROW ECHELON FORMAT
    
    for i in range(N-2,-1,-1):
        for j in range(N-1,i,-1):
            if ( A[i,j] != 0.0 ):
                X[i] += - A[i,j] * X[j]
                A[i,:] += - A[i,j] * A[j,:]
    
    #---------------------------------------
    
    return X

def _CreateTracks(Mstar,Age,dAge,OmegaEnv,OmegaCore,params,StarEvo):
    """Takes simulation parameters, returns dictionary with empty arrays for evolutionary tracks."""
    
    # Create empty dictionary
    Tracks = {}
    
    # Add basic tracks that are always included
    Tracks['Age'] = np.array([])
    Tracks['dAge'] = np.array([])
    Tracks['OmegaEnv'] = np.array([])
    Tracks['OmegaCore'] = np.array([])
    
    # If doing extended tracks, add these
    if params['ExtendedTracks']:
  
        # Add extended activity quantities to dictionary
        # note: do not need to call phys.RotationQuantities since without StarState being passed as an argument, it will be called 
        #       by ExtendedQuantities() anyway.
        StarState = phys.ExtendedQuantities( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , OmegaCore=OmegaCore , params=params , StarEvo=StarEvo )
        
        # Add empty arrays for each of these values 
        for quantity in StarState:
            Tracks[quantity] = np.array([])
  
    else:
        
        # Just set StarState to None since it is not needed
        StarState = None
    
    # Add initial values to tracks
    Tracks = _AppendTracks(Tracks,Mstar,Age,dAge,OmegaEnv,OmegaCore,params,StarEvo,StarState=StarState)
    
    return Tracks
  
def _AppendTracks(Tracks,Mstar,Age,dAge,OmegaEnv,OmegaCore,params,StarEvo,StarState=None):
    """Takes state of evolutionary simulation, appends state onto arrays in Tracks dictionary."""
    
    # Basic quantities that are always included
    Tracks['Age'] = np.append( Tracks['Age'] , Age )
    Tracks['dAge'] = np.append( Tracks['dAge'] , 0.0 )
    Tracks['OmegaEnv'] = np.append( Tracks['OmegaEnv'] , OmegaEnv )
    Tracks['OmegaCore'] = np.append( Tracks['OmegaCore'] , OmegaCore )
    
    # If doing extended tracks, calculate and add them
    if params['ExtendedTracks']:
        
        # Calculate extended quantities if not input as keyword argument
        if StarState is None:
            
            # note: do not need to call phys.RotationQuantities since without StarState being passed as an argument, it will be called 
            #       by ExtendedQuantities() anyway.
            StarState = phys.ExtendedQuantities( Mstar=Mstar , Age=Age , OmegaEnv=OmegaEnv , OmegaCore=OmegaCore , params=params , StarEvo=StarEvo )
            
        # Append these quantities, but only if they are not already present
        for quantity in StarState:
            
            # Do not do quantity if it was already added above
            if not ( quantity in ['Age','dAge','OmegaEnv','OmegaCore'] ):
                Tracks[quantity] = np.append( Tracks[quantity] , StarState[quantity] )
        
    return Tracks

def _CheckBadData(Tracks):
    """Takes current set of tracks, checks last entries for bad data."""
    
    # Originally assume no error and sould continue
    error = False
    
    # Make sure rotation rates are positive
    if ( Tracks['OmegaEnv'][-1] <= 0.0 ):
        misc._PrintError("negative value of OmegaEnv in rotational evolution calculation")
        error = True
    if ( Tracks['OmegaCore'][-1] <= 0.0 ):
        misc._PrintError("negative value of OmegaCore in rotational evolution calculation")
        error = True
        
    # Loop over elements in Tracks and check each one for NaN or Inf
    for track in Tracks:
        if math.isnan(Tracks[track][-1]):
            misc._PrintError("NaN found for "+track+" in rotational evolution calculation")
            error = True
        if math.isinf(Tracks[track][-1]):
            misc._PrintError("inf found for "+track+" in rotational evolution calculation")
            error = True
    
    # If there was an error, kill the code
    if error:
        misc._PrintErrorKill("ending simulation due to bad data found")
    
    return



"""Module holding the Cluster class and related functions."""

# Imports for standard stuff needed here
import sys
import inspect
import numpy as np
import os
import pickle

# Imports for mors modules
import mors.miscellaneous as misc
import mors.stellarevo as SE
import mors.parameters as params
import mors.rotevo as RE
import mors.physicalmodel as phys
import mors.star as star

class Cluster:
    """A class for star objects that hold all information about a star. 
    
    Attributes
    ------------
    
    Methods
    ------------
    
    """
        
    def __init__(self,Mstar=None,Age=None,Omega=None,OmegaEnv=None,OmegaCore=None,AgesOut=None,starEvoDir=None,evoModels=None,params=params.paramsDefault,verbose=False):
        """Initialises instance of Cluster class.
        
        This is the main function that is run when creating an instance of the Cluster class and it sets up all
        the things needed including calculating evolutionary tracks for each of the stars star. The function requires 
        that an array of stellar masses (in Msun) and initial rotation rates (in OmegaSun=2.67e-6 rad s^-1) are input.
        using the Mstar and Omega keyword arguments. Alternatively, OmegaEnv and OmegaCore can be set, in which case
        Omega does not need to be specified. If the argument Age (in Myr) is also specified, then the code will find 
        the evolutionary tracks for each star that passes through this rotation rate at this age, otherwise if Age is 
        not set then it will calculate evolutionary tracks assuming this Omega as the initial (1 Myr) rotation rate. 
        The user should not specify OmegaCore and Age simultaneously, and if Age is set then either Omega or OmegaEnv 
        can be used to specify the surface rotation rate. Age can also be set as an array giving different ages for 
        each star.
        
        """
        
        # Make sure Mstar was set
        if Mstar is None:
            misc._PrintErrorKill("Mstar keyword argument not set")
        
        # Make sure Omega, OmegaEnv, and OmegaCore are well set
        # (if Omega is set, OmegaEnv and OmegaCore will both be set to that value)
        Omega , OmegaEnv , OmegaCore = _CheckInputRotation(Age,Omega,OmegaEnv,OmegaCore)
        
        # Make sure arrays all same length
        if not ( len(Mstar) == len(Omega) ):
            misc._PrintErrorKill("Mstar and Omega have different lengths")
        if not ( len(Mstar) == len(OmegaEnv) ):
            misc._PrintErrorKill("Mstar and OmegaEnv have different lengths")
        if not ( len(Mstar) == len(OmegaCore) ):
            misc._PrintErrorKill("Mstar and OmegaCore have different lengths")
        
        # Set number of stars
        self.nStars = len(Mstar)
        
        # Set parameters
        self.params = params
        
        # Set the ExtendedTracks parameter to True so we get all parameters
        self.params['ExtendedTracks'] = True
        
        # Set Mstar arrays
        self.Mstar = Mstar
        
        # Set AgesOut 
        self.AgesOut = AgesOut
        
        # Set stellar evo model
        self.starEvoDir = starEvoDir
        self.evoModels = evoModels
        
        # Get evolutionary tracks 
        self._LoadEvoTracks(Age,OmegaEnv,OmegaCore,verbose)
        
        # Get HZ boundaries
        self.aOrbHZ = phys.aOrbHZ(Mstar=self.Mstar,params=self.params)
        
        return
        
    def _LoadEvoTracks(self,Age,OmegaEnv0,OmegaCore0,verbose):
        """Loads rotation and activity tracks for each star."""
        
        # Start dictionary holding stars
        self.stars = []
        
        # Loop over each star and load each one
        for iStar in range(0,len(self.Mstar)):
            
            # Print to screen if verbose was set by user
            if verbose:
                print("Loading star "+str(iStar)+" out of "+str(self.nStars)+" in cluster",end="\r")
            
            # Get parameters for this star in right type
            MstarStar = float(self.Mstar[iStar])
            OmegaEnvStar = float(OmegaEnv0[iStar])
            OmegaCoreStar = float(OmegaCore0[iStar])
            
            # Get Age depending on if it is an array or a float/int
            if Age is None:
                AgeIn = None
            elif isinstance(Age,(float,int)):
                AgeIn = Age
            else:
                AgeIn = Age[iStar]
            
            # Create instance of star class for this star
            starTemp = star.Star(Mstar=MstarStar,Age=AgeIn,OmegaEnv=OmegaEnvStar,OmegaCore=OmegaCoreStar,
                                AgesOut=self.AgesOut,starEvoDir=self.starEvoDir,evoModels=self.evoModels,params=self.params)
            
            # Append dictionary
            self.stars.append(starTemp)
            
            # Also make this star an attribute of the class
            setattr( self , "star"+str(iStar) , starTemp )
            
        # Make new line if printing status to screen
        if verbose:
            print("")
        
        # Make functions for each quantity that return this quantity at a given age as attributes of class
        self._setupQuantityFunctions()
        
        return
        
    def _setupQuantityFunctions(self):
        """Makes functions for each quantity that return this quantity at a given age as attributes of class."""
        
        # A description of how the code below works is given in the star class function with the same name
        # as this function. In this case, the code does it based on the quantities held in the first star. 
        for track in self.stars[0].Tracks:
        
            exec( "def "+track+"(self,Age):\n  return self.Values(Age=Age,Quantity='"+track+"')" )
            exec( "setattr( self.__class__ , '"+track+"' , "+track+" )" )
        
        return
    
    def Values(self,Age=None,Quantity=None):
        """Takes age in Myr and a string with name of quantity to output, returns value of that quantity at specified age for all stars.
        
        Parameters
        ----------
        Age : float
            Age of cluster.
        Quantity : str
            String holding name of parameter to get.
        
        Returns
        ----------
        values : numpy.ndarray
            Set of extended quantities.
        
        """
        
        # Make sure input parameters are set
        if Age is None:
            misc._PrintErrorKill("keyword parameter Age not set in call to function")
        if Quantity is None:
            misc._PrintErrorKill("keyword parameter Quantity not set in call to function")
        
        # Make array to hold values
        values = np.zeros(self.nStars)
        
        # Loop over stars and get values for each
        for iStar in range(0,self.nStars):
            values[iStar] = self.stars[iStar].Value(Age=Age,Quantity=Quantity)
        
        return values
        
    def PrintStars(self):
        """Prints list of stars in cluster to screen."""
        
        # Header
        print("The following is a list of masses for stars in this cluster.")
        
        # Loop over each star and print basic parameters
        for iStar in range(0,self.nStars):
            print("   "+str(iStar)+". "+str(self.Mstar[iStar])+" Msun")
        
        # Total number of stars
        print("Number of stars in cluster = "+str(self.nStars))
        
        return
    
    def Save(self,filename='cluster.pickle'):
        """Takes filename (default is 'cluster.pickle'), saves cluster to this file using pickle."""
        
        with open(filename,'wb') as f:
            pickle.dump(self,f)
        
        return
    
    def save(self,filename='cluster.pickle'):
        """Same as Save()."""
        self.Save(filename=filename)
        return    
        
    def Percentile(self,Mstar=None,Age=None,Omega=None,Prot=None,percentile=None):
        """Gets rotation rate of percentile or percentile of rotation rate in the rotation distribution at given age.
        
        This function can be used for two purposes
        1. to determine the percentile in a rotation distribution of a star given its mass, rotation rate, and age
        2. to determine the rotation rate of a star in a rotation distribution given its mass, percentile, and age
        In the first case, the user should specify the rotation rate using either the Omega or Prot keyword arguments (given 
        in OmegaSun and days respectively) and the mass. In the second case, the user should specify the percentile using the
        percentile keyword argument (in the range 0 to 100) and the mass. The mass should be specified in Msun using the Mstar
        keyword argument and the age in Myr using the Age keyword argument. These should only be given as floats. 
        
        Parameters
        ----------
        Mstar : float
            Stellar mass in Msun.
        Age : float
            Age of star in Myr.
        Omega : float , optional
            Rotation rate of star in OmegaSun.
        Prot : float , optional
            Rotation period of star in days.
        percentile : float , optional
            percentile in distribution (between 0 and 100).
        
        Returns
        ----------
        result : float
            Either rotation rate for percentile or percentile for rotation rate.
        
        """
        
        # Make sure the age is specified
        if Age is None:
            misc._PrintErrorKill("keyword parameter Age not set in call to function")
        
        # If percentile was set to a string, get float version
        if isinstance(percentile,str):
            if ( percentile == 'slow' ):
                percentile = 5.0
            elif ( percentile == 'medium' ):
                percentile = 50.0
            elif ( percentile == 'fast' ):
                percentile = 95.0
            else:
                misc._PrintErrorKill( "invalid percentile string (options are 'slow', 'medium', or 'fast')" )
        
        # Get the rotation distribution at this age
        OmegaDist = self.Values( Age=Age , Quantity='OmegaEnv' )
        
        # Get the result
        result = star.Percentile( Mstar=Mstar , Omega=Omega , Prot=Prot , percentile=percentile , 
                                MstarDist=self.Mstar , OmegaDist=OmegaDist , params=self.params )
        
        return result
    
    def ActivityLifetime(self,Quantity=None,Threshold=None,AgeMax=None):
        """Takes threshold value, returns ages at which each star last drops below this threshold.
        
        This function can be used to determine when each star's emission crosses a given threshold value for a few
        activity quantities. These are Lx, Fx, Rx, and FxHZ for X-rays, and similar values for EUV1, EUV2, EUV, 
        XUV, and Ly-alpha. If the star crosses the threshold (from above it to below it) multiple times, this 
        will find the final time it will cross the threshold. If the user wants to set a maximum age so that the 
        code only looks for crossings of the threshold below this age then this can be done using the AgeMax
        keyword argument.
            
        Parameters
        ----------
        Quantity : str
            Gives which quantity to consider (e.g. 'Lx').
        Threshold : float or str
            Value for threshold in units of quantity or string of 'sat'.
        AgeMax : float , optional
            End age of track to consider in Myr.
        
        Returns
        ----------
        AgeActive : float
            Activity lifetime in Myr.
        
        """
        
        # Make sure Quantity is set
        if Quantity is None:
            misc._PrintErrorKill("Quantity not set in call to function")
        
        # Make sure Quantity is string
        if not isinstance(Quantity,str):
            misc._PrintErrorKill("Quantity must be string")
        
        # Make array
        AgeActive = np.zeros(self.nStars)
        
        # Loop over stars and get each
        for iStar in range(self.nStars):
            AgeActive[iStar] = self.stars[iStar].ActivityLifetime(Quantity=Quantity,Threshold=Threshold,AgeMax=AgeMax)
        
        return AgeActive
        
    def IntegrateEmission(self,AgeMin=None,AgeMax=None,Band=None,aOrb=None):
        """Takes age range, returns integrated emission in band within that range for each star.
        
        This code can be used to the luminosities of the stars between two ages. This can be applied to any wavelength band
        and the result is a total energy emitted in this time in erg. If the user also specifies an orbital distance using 
        the aOrb keyword argument, the code integrates the flux at this obital distance returns a fluence in erg cm^-2. The
        user can specify aOrb as a string to get the fluences at various habitable zone boundaries (using the HZ calculated 
        at the age defined in params when creating this cluster). Options are 'RecentVenus', 'RunawayGreenhouse', 'MoistGreenhouse', 
        'MaximumGreenhouse', 'EarlyMars', and 'HZ'.
            
        Parameters
        ----------
        AgeMin : float
            Start of time period to integrate in Myr.
        AgeMax : float
            End of time period to integrate in Myr.
        Band : str
            Gives which wavelength band to consider (options are 'XUV', 'Xray', 'EUV1', 'EUV2', 'EUV', 'Lyman', 'bol').
        aOrb : float or str, optional
            Orbital distance to get fluence at in AU or string identifying HZ boundary.
        
        Returns
        ----------
        Energy : numpy.ndarray
            Integrated luminosity or flux in erg or erg cm^-2.
        
        """
        
        # Make array
        Energy = np.zeros(self.nStars)
        
        # Loop over stars and get each
        for iStar in range(self.nStars):
            Energy[iStar] = self.stars[iStar].IntegrateEmission(AgeMin=AgeMin,AgeMax=AgeMax,Band=Band,aOrb=aOrb)
        
        return Energy

def _CheckInputRotation(Age,Omega,OmegaEnv,OmegaCore):
  """Takes input rotation, checks if values are setup correctly."""
  
  # Make sure if Age is set that OmegaCore is not set
  if ( not Age is None ) and ( not OmegaCore is None ):
    misc._PrintErrorKill( "cannot set both Age and OmegaCore as arguments of Cluster" )
  
  # Make sure at least one rotation rate is set
  if ( Omega is None ):
    
    # Since Omega is not set, make sure both OmegaEnv and OmegaCore are set
    if ( OmegaEnv is None ) or ( OmegaCore is None ):
      misc._PrintErrorKill( "must set either Omega or both OmegaEnv and OmegaCore as arugment of Star" )
  
  else:
    
    # Since Omega is set, make sure both OmegaEnv and OmegaCore are not set
    if not ( ( OmegaEnv is None ) and ( OmegaCore is None ) ):
      misc._PrintErrorKill( "cannot set OmegaEnv and OmegaCore when Omega is set as arugment of Star" )
    
    # Set both OmegaEnv and OmegaCore to Omega
    OmegaEnv = Omega
    OmegaCore = Omega
  
  return Omega , OmegaEnv , OmegaCore




"""Module for loading the stellar evolution tracks and retrieving basic stellar properties."""

# Imports for standard stuff needed here
import numpy as np
import pickle
import os
import math
import sys
import copy
import inspect

# Imports for mors modules
import mors.miscellaneous as misc

#----------------------------------------------------------
# Parameters for stellar evolution models

# Directory for stellar evolution models
fwl_data_dir = os.getenv('FWL_DATA')
starEvoDirDefault = fwl_data_dir + "/stellar_evolution_tracks/Spada/fs255_grid/"

# Set which set of models to use, i.e. which X, Z, and A values (set to X0p70952_Z0p01631_A1p875 for closest to solar)
evoModelsDefault = "X0p70952_Z0p01631_A1p875"

# Initially set default model data to None so we know it has not been set
ModelDataDefault = None

# Number of decimal places to represent masses and ages when doing interpolations
nDecValue = 7

#----------------------------------------------------------

class StarEvo:
    """Class to hold complete information about stellar evolution models.
    
    An instance of this class will hold evolutionary tracks for all masses and all quantities of interest. For
    now, only models from Spada et al. (2013) can be loaded. The main set of information held in this class is
    in ModelData, which is a dictionary of dictionaries. Each dictionary in this dictionary corresponds to a 
    single stellar mass and holds evolutionary tracks for a bunch of important quantities.

    Attributes
    ------------
    ModelData : dict
        This is a dictionary of dictionaries, holding evolution tracks for each stellar mass.
    
    Methods
    ------------
    
    """
        
    def __init__(self,starEvoDir=starEvoDirDefault,evoModels=evoModelsDefault):
        """Initialises instance of StarEvo class."""
        
        # If starEvoDir and evoModels are None, use defaults
        if starEvoDir is None:
            starEvoDir = starEvoDirDefault
        if evoModels is None:
            evoModels = evoModelsDefault
        
        # Load the models
        self.ModelData = _LoadModels( starEvoDir=starEvoDir , evoModels=evoModels )
        
        return
        
    def LoadTrack(self,Mstar,ClearData=False):
        """Takes stellar mass, loads evolutionary track for a specific mass into the model data.
        
        This can be useful if the user wants to load a specific evolutionary track into a model data for a specific mass
        since this will make getting values for this mass faster since no interpolation between mass bins will be needed.
        
        Parameters
        ----------
        Mstar : float
            Mass of star in Msun.
        ClearData : bool , optional
            If set to True, all other stellar masses will be removed from the ModelData dictionary.
        
        Returns
        ----------
        None
            None
        
        """
        
        self.ModelData = LoadTrack(Mstar,ModelData=self.ModelData,ClearData=ClearData)
        
        return
        
    def Value(Mstar,Age,ParamString):
        """Takes stellar mass, age, and a parameter string, returns values corresponding to named parameter.
        
        The set of models should have already been loaded. With this function, the user can ask for a value
        of one of the parameters for a specific stellar mass and age. All three of these can be input as multiple
        values and an array of values will be returned if this is the case. For example, if MstarIn is input as 
        a 1D array, the function will return a 1D array giving the value for each of these masses. ParamString
        can be input as a list of strings and values for each parameter in that list will be returned in a 1D 
        array. If two are given as arrays or lists, then a 2D array will be returned. If all three then a 3D 
        array with dimensions len(MstarIn)xlen(AgeIn)xlen(ParamString) will be returned.
        
        Parameters
        ----------
        MstarIn : float or int or numpy.ndarray
            Mass of star in Msun.
        AgeIn : float or int or numpy.ndarray
            Age in Myr.
        ParamString : str
            String holding name of parameter to get value for.
        
        Returns
        ----------
        value : float or numpy.ndarray
            Value of parameter at this mass and age. 
        
        """
        
        # Call Value() outside this class to get the value with this model
        value = Value(Mstar,Age,ParamString,ModelData=self.ModelData)
        
        return
    
    # The following functions are for individual parameters that can be called
    def Rstar(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns stellar radius."""
        return Value( Mstar , Age , 'Rstar' , ModelData=self.ModelData )

    def Lbol(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns bolometric luminosity."""
        return Value( Mstar , Age , 'Lbol' , ModelData=self.ModelData )
    
    def Teff(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns effective temperature."""
        return Value( Mstar , Age , 'Teff' , ModelData=self.ModelData )
    
    def Itotal(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns total moment of interia."""
        return Value( Mstar , Age , 'Itotal' , ModelData=self.ModelData )
    
    def Icore(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns core moment of interia."""
        return Value( Mstar , Age , 'Icore' , ModelData=self.ModelData )
    
    def Ienv(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns envelope moment of interia."""
        return Value( Mstar , Age , 'Ienv' , ModelData=self.ModelData )
    
    def Mcore(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns core mass."""
        return Value( Mstar , Age , 'Mcore' , ModelData=self.ModelData )
    
    def Menv(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns envelope mass."""
        return Value( Mstar , Age , 'Menv' , ModelData=self.ModelData )
    
    def Rcore(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns core radius."""
        return Value( Mstar , Age , 'Rcore' , ModelData=self.ModelData )
    
    def tauConv(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns convective turnover time."""
        return Value( Mstar , Age , 'tauConv' , ModelData=self.ModelData )
    
    def dItotaldt(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns rate of change of total moment of inertia."""
        return Value( Mstar , Age , 'dItotaldt' , ModelData=self.ModelData )
    
    def dIcoredt(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns rate of change of core moment of inertia."""
        return Value( Mstar , Age , 'dIcoredt' , ModelData=self.ModelData )
    
    def dIenvdt(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns rate of change of envelope moment of inertia."""
        return Value( Mstar , Age , 'dIenvdt' , ModelData=self.ModelData )
    
    def dMcoredt(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns rate of change of core mass."""
        return Value( Mstar , Age , 'dMcoredt' , ModelData=self.ModelData )
    
    def dRcoredt(self,Mstar,Age,ModelData=ModelDataDefault):
        """Takes mass and age, returns rate of change of core radius."""
        return Value( Mstar , Age , 'dRcoredt' , ModelData=self.ModelData )
    
def _LoadModels(starEvoDir=starEvoDirDefault,evoModels=evoModelsDefault):
    """Loads evolutionary tracks as a grid of parameters at each mass and age."""
    
    # Check if should compile new grid of evolutionary models or load previous grid
    if _shouldCompileNew(starEvoDir,evoModels):
        ModelData = _CompileNewGrid(starEvoDir,evoModels)
    else:
        ModelData = _LoadSavedGrid(starEvoDir,evoModels)
    
    return ModelData

def _shouldCompileNew(starEvoDir,evoModels):
    """Takes directory for stellar evo models, returns if new grid needs to be compiled."""
    
    # Initially assume need to compile new
    compileNew = True
    
    # Check if previously compiled models already exist
    if ( os.path.isfile(starEvoDir+evoModels+".pickle") ):
        compileNew = False
    
    return compileNew

def _CompileNewGrid(starEvoDir,evoModels):
    """Compiles new dictionary of stellar evo models."""
    
    # List all masses to load
    MstarAll = np.array([ 0.1 , 0.15 , 0.2 , 0.25 , 0.3 , 0.35 , 0.4 , 0.45 , 0.5 , 0.55 , 0.6 , 0.65 , 0.7 , 0.75 , 0.8 , 0.85 , 0.9 , 0.95 , 1.0 , 1.05 , 1.1 , 1.15 , 1.2 , 1.25 ])
    
    # Mass components of the filenames for these masses
    MstarFilenameMiddle = [ '0p10' , '0p15' , '0p20' , '0p25' , '0p30' , '0p35' , '0p40' , '0p45' , '0p50' , '0p55' , '0p60' , '0p65' , '0p70' , '0p75' , '0p80' , '0p85' , '0p90' , '0p95' , '1p00' , '1p05' , '1p10' , '1p15' , '1p20' , '1p25' ]
    
    # Start empty dictionary
    ModelData = {}
    
    # Add stellar mass array 
    ModelData['MstarAll'] = MstarAll
    
    # Add a list of strings holding each of the parameters
    ModelData['ParamsAll'] = [ 'Mstar' , 'Age' , 'Rstar' , 'Lbol' , 'Teff' , 'Itotal' , 'Icore' , 'Ienv' , 'Mcore' , 'Menv' , 'Rcore' , 
                                'tauConv' , 'dItotaldt' , 'dIcoredt' , 'dIenvdt' , 'dMcoredt' , 'dRcoredt' ]
    
    # Loop over masses and add each one to dictionary
    for iMstar in range(0,len(MstarAll)):
        ModelData[MstarAll[iMstar]] = _ReadEvolutionTrack( starEvoDir , evoModels , MstarAll[iMstar] , MstarFilenameMiddle[iMstar] )
    
    # Save compiled models
    with open(starEvoDir+evoModels+".pickle",'wb') as f:
        pickle.dump(ModelData,f)
        
    
    return ModelData

def _ReadEvolutionTrack(starEvoDir,evoModels,Mstar,MstarFilenameMiddle):
    """Loads the stellar evolution models from Spada et al. (2013) for a mass bin and puts it into a dictionary."""
        
    # Set strings for starting and ending of filenames
    filename_prefix = starEvoDir + evoModels + "/M"
    filename_postfix1 = "_" + evoModels + ".track1"
    filename_postfix2 = "_" + evoModels + ".track2"
    
    # Get names of data files holding evo models
    filename1 = filename_prefix + MstarFilenameMiddle + filename_postfix1
    filename2 = filename_prefix + MstarFilenameMiddle + filename_postfix2
    
    # Read contents of files
    with open(filename1,'r') as f:
        content1 = f.readlines()
    with open(filename2,'r') as f:
        content2 = f.readlines()
    
    # Number of header lines
    nHeader = 1
    
    # Remove any duplicate ages
    content1 = _RemoveDuplicateAges(nHeader,content1)
    content2 = _RemoveDuplicateAges(nHeader,content2)
    
    # Make sure they are the same lengths
    if not ( len(content1) == len(content2) ):
        misc._PrintErrorKill("two evo files not same length")
    
    # Get number of age bins
    nAge = len(content1) - nHeader
        
    # Make arrays to hold desired quantities
    Age = np.zeros(nAge)
    Rstar = np.zeros(nAge)
    Lbol = np.zeros(nAge)
    Teff = np.zeros(nAge)
    Itotal = np.zeros(nAge)
    Icore = np.zeros(nAge)
    Ienv = np.zeros(nAge)
    Mcore = np.zeros(nAge)
    Menv = np.zeros(nAge)
    Rcore = np.zeros(nAge)
    tauConv = np.zeros(nAge)
    dItotaldt = np.zeros(nAge)
    dIcoredt = np.zeros(nAge)
    dIenvdt = np.zeros(nAge)
    dMcoredt = np.zeros(nAge)
    dRcoredt = np.zeros(nAge)
        
    # Read quantities from first file into the arrays (has age, Lbol, and Rstar)
    iAge = 0
    for line in content1[nHeader:len(content1)]:
        
        # Split data into list
        data = line.split()
        
        # Read data 
        Age[iAge] = data[0]
        Lbol[iAge] = data[4]
        Teff[iAge] = data[7]
        Rstar[iAge] = data[5]
        
        # Update index
        iAge += 1
        
    # Read quantities from second file into the arrays (has Itotal, Ienv, Menv, Rcore, and tauConv)
    iAge = 0
    for line in content2[nHeader:len(content2)]:
        
        # Split data into list
        data = line.split()
        
        # Read data 
        Itotal[iAge] = data[7]
        Ienv[iAge] = data[6]
        Menv[iAge] = data[2]
        Rcore[iAge] = data[3]
        tauConv[iAge] = data[4]
        
        # Update i
        iAge += 1
        
    # Change some units and convert from log to lin
    Age[:] = Age[:] * 1000.0 # convert from Gyr into Myr
    Rstar[:] = 10.0**Rstar[:] # convert into linear since Spada has in log
    Lbol[:] = 10.0**Lbol[:] # also this one
    Teff[:] = 10.0**Teff[:] # also this one
        
    # Get some core quantities
    Icore[:] = Itotal[:] - Ienv[:]
    Mcore[:] = ( 1.0 - Menv[:] ) * Mstar
    Rcore[:] = Rcore[:] * Rstar[:]
        
    # Get gradients
    dItotaldt[:] = _CalculateGradient( Age , Itotal )
    dIcoredt[:] = _CalculateGradient( Age , Icore )
    dIenvdt[:] = _CalculateGradient( Age , Ienv )
    dMcoredt[:] = _CalculateGradient( Age , Mcore )
    dRcoredt[:] = _CalculateGradient( Age , Rcore )
    
    # Round mass and ages to nDecValue decimal places as specified at top of file
    Mstar = round( Mstar , nDecValue )
    Age = np.round( Age , decimals=nDecValue )
    
    # Add all quantities to the dictionary
    Data = {}
    Data['Mstar'] = Mstar
    Data['Age'] = Age
    Data['Rstar'] = Rstar
    Data['Lbol'] = Lbol
    Data['Teff'] = Teff
    Data['Itotal'] = Itotal
    Data['Icore'] = Icore
    Data['Ienv'] = Ienv
    Data['Mcore'] = Mcore
    Data['Menv'] = Menv
    Data['Rcore'] = Rcore
    Data['tauConv'] = tauConv
    Data['dItotaldt'] = dItotaldt
    Data['dIcoredt'] = dIcoredt
    Data['dIenvdt'] = dIenvdt
    Data['dMcoredt'] = dMcoredt
    Data['dRcoredt'] = dRcoredt
    
    return Data

def _RemoveDuplicateAges(nHeader,content):
    """Takes lines from evo file, removes duplicate ages."""
    
    # This is necessary because one of the Spada files has an age repeated a few times.
    
    # Start list
    contentNew = []
    
    # Add header lines
    for line in content[0:nHeader]:
        contentNew.append(line)
    
    # Get age of first line
    data = content[nHeader].split()
    AgeLast = float(data[0])
    
    # Loop over remaning lines and decide which to add (all but duplicate ages)
    for line in content[nHeader+1:len(content)]:
        
        # Get age
        data = line.split()
        Age = float(data[0])
        
        # Add line if not same age (within some tolerance)
        if not ( abs(Age/AgeLast-1.0) < 1.0e-5 ):
            contentNew.append(line)
        
        # Update AgeLast
        AgeLast = Age
    
    return contentNew

def _CalculateGradient(Age,X):
    """Takes an evolutionary track for quantity, returns evolution of rate of change of quantity."""
    
    # Number of age bins
    nAge = len(Age)
    
    # Make array
    dXdt = np.zeros(nAge)
    
    # Set first and last elements
    dXdt[0] = ( X[1] - X[0] ) / ( Age[1] - Age[0] )
    dXdt[nAge-1] = ( X[nAge-1] - X[nAge-2] ) / ( Age[nAge-1] - Age[nAge-2] )
    
    # Do rest of ages
    dXdt[1:nAge-1] = ( X[2:nAge] - X[0:nAge-2] ) / ( Age[2:nAge] - Age[0:nAge-2] )
    
    dAge = np.zeros(nAge)
    dAge[0] = Age[1] - Age[0]
    dAge[1:nAge-1] = Age[2:nAge] - Age[0:nAge-2]
    dAge[nAge-1] = Age[nAge-1] - Age[nAge-2]
    
    return dXdt

def _LoadSavedGrid(starEvoDir,evoModels):
    """Takes filename for stellar evo model, returns grid of models."""
    
    # Simply load data
    with open(starEvoDir+evoModels+".pickle",'rb') as f:
        ModelData = pickle.load(f)
        
    return ModelData

def LoadTrack(Mstar,ModelData=None,ClearData=False):
    global ModelDataDefault
    """Takes stellar mass, loads evolutionary track for a specific mass into the model data.
    
    This can be useful if the user wants to load a specific evolutionary track into a model data for a specific mass
    since this will make getting values for this mass faster since no interpolation between mass bins will be needed.
    The user might also want to set ClearData to True if doing the calculation only for a single stellar mass since 
    this will cause the function to remove data for all other masses, which saves memory, though usually this will 
    only be necessary if the user is going to have data for a very large number of stars loaded simultaneously. Note
    that this does not change ModelDataDefault.
        
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    ModelData : dict , optional
        Dictionary of dictionaries holding set of stellar evolution models.
    ClearData : bool , optional
        If set to True, all other stellar masses will be removed from the ModelData dictionary.
    
    Returns
    ----------
    None
        None
    """
    
    # Use default model data if nothing was specified
    if ModelData is None:
        ModelData = ModelDataDefault

    # Check if ModelData is not None, otherwise need to load defaults
    if ModelData is None:
        ModelData = _LoadDefaultModelData()
        
    # Only do something if Mstar is not already in ModelData
    if not ( Mstar in ModelData ):
        
        # Get evo tracks for this mass and add it to the dictionary
        Data = _LoadTrack(Mstar,ModelData)
        
        # Add extra term to ModelData
        ModelData[Mstar] = Data
        
        # Add this mass to MstarAll in ModelData
        ModelData['MstarAll'] = np.sort( np.append( ModelData['MstarAll'] , Mstar ) ) 
        
        # If ModelDataDefault is not None, add this also to that
        if not ModelDataDefault is None:
            ModelDataDefault[Mstar] = Data
        
    # If ClearData is set, remove all other tracks from return dictionary, but not ModelDataDefault
    if ClearData:
        
        # Do this by making a new dictionary and copying stuff in
        
        # Make deepcopy of dictionary
        ModelDataOld = copy.deepcopy(ModelData)
        
        # Make new dictionary
        ModelData = {}
        
        # Make new MstarAll array
        ModelData['MstarAll'] = np.array([Mstar])
        
        # Add properties
        ModelData['ParamsAll'] = ModelDataOld['ParamsAll']
        
        # Add tracks for this mass bin
        ModelData[Mstar] = ModelDataOld[Mstar]
    
    return ModelData

def _LoadTrack(Mstar,ModelData):
    """Takes stellar mass and model data dictionary, returns dictionary with track for this mass."""
    
    # Round mass to nDecValue decimal places specified at top of this file
    Mstar = round( Mstar , nDecValue )
    
    # Make sure within mass limit
    _CheckMassLimit( ModelData['MstarAll'] , Mstar )
    
    # Get nearest mass bin below
    iMin = misc._getIndexLT( ModelData['MstarAll'] , Mstar )
    MstarMin = ModelData['MstarAll'][iMin]
    
    # Get nearest mass bin above
    iMax = misc._getIndexGT( ModelData['MstarAll'] , Mstar )
    MstarMax = ModelData['MstarAll'][iMax]
    
    # Get dictionaries for masses below and above Mstar
    DataMin = ModelData[MstarMin]
    DataMax = ModelData[MstarMax]
    
    # Get min and max ages to include
    #   for min, this is the maximum starting age of the two tracks
    #   for max, this is the minimum ending age of the two tracks
    AgeMin = np.max( [ DataMin['Age'][0] , DataMax['Age'][0] ] )
    AgeMax = np.min( [ DataMin['Age'][-1] , DataMax['Age'][-1] ] )
    
    # Number of age bins to use, just make this minimum of the number in the two tracks
    nAge = np.min( [ len(DataMin['Age']) , len(DataMax['Age']) ] )
    
    # Make age array, log spacing
    Age = np.logspace( np.log10(AgeMin) , np.log10(AgeMax) , nAge )
    
    # Round ages to nDecValue decimal placed specified at top of this file
    Age = np.round( Age , decimals=nDecValue )
    
    # Start the output dictionary
    Data = {}
    
    # Add mass and age array to dictionary
    Data['Mstar'] = Mstar
    Data['Age'] = Age
    
    # Loop over parameters and get tracks for each (do parameters in DataMin)
    for param in DataMin:
        
        # Skip parameter if already added (so don't do Mstar and Age again)
        if ( param in Data ):
            continue
        
        # Get track
        track = Value( Mstar , Age , param , ModelData=ModelData )
        
        # Add to dictionary
        Data[param] = track
    
    return Data

def _LoadDefaultModelData():
    global ModelDataDefault
    """Loads model data for the default model to be used if no StarEvo class object is used."""
    
    # Loading a default model is generally not necessary, but it would be useful if the user doesn't want to mess
    # around with creating a StarEvo class object and instead just wants to quickly get a few values.
    
    # Load the data 
    ModelDataDefault = _LoadModels()
    
    return ModelDataDefault

def Value(MstarIn,AgeIn,ParamString,ModelData=ModelDataDefault):
    """Takes stellar mass, age, and a parameter string, returns values corresponding to named parameter.
    
    The set of models should have already been loaded. With this function, the user can ask for a value
    of one of the parameters for a specific stellar mass and age. All three of these can be input as multiple
    values and an array of values will be returned if this is the case. For example, if MstarIn is input as 
    a 1D array, the function will return a 1D array giving the value for each of these masses. ParamString
    can be input as a list of strings and values for each parameter in that list will be returned in a 1D 
    array. If two are given as arrays or lists, then a 2D array will be returned. If all three then a 3D 
    array with dimensions len(MstarIn)xlen(AgeIn)xlen(ParamString) will be returned.
    
    Parameters
    ----------
    MstarIn : float or int or numpy.ndarray
        Mass of star in Msun.
    AgeIn : float or int or numpy.ndarray
        Age in Myr.
    ParamString : str
        String holding name of parameter to get value for.
    ModelData : dict , optional
        Dictionary of dictionaries holding set of stellar evolution models.
    
    Returns
    ----------
    value : float or numpy.ndarray
        Value of parameter at this mass and age. 
    
    """
    
    # Check if ModelData is not None, otherwise need to load defaults
    if ModelData is None:
        ModelData = _LoadDefaultModelData()
    
    # There are now eight scenarios here
    #   1. Mstar and Age are both floats, so return float.
    #   2. Mstar is float and Age is numpy.ndarray, so return numpy.ndarray of length len(Age).
    #   3. Mstar is numpy.ndarray and Age is float, so return numpy.ndarray of length len(Age)
    #   4. Mstar and Age are both numpy.ndarray, so return numpy.ndarray of shape ( len(Mstar) , len(Age) ).
    # Note: could use type(X).__module__ == 'numpy' to test for numpy but it doesn't work if 
        
    # Get Mstar and Age in correct types
    Mstar = misc._convertFloatArray(MstarIn)
    Age = misc._convertFloatArray(AgeIn)
    
    # Make sure Mstar and Age are correct types
    if Mstar is None:
        misc._PrintErrorKill("argument Mstar has invalid type")
    if Age is None:
        misc._PrintErrorKill("argument Age has invalid type")
        
    # Find if scenario 1 (most likely)
    if ( isinstance(Mstar,float) and isinstance(Age,float) and isinstance(ParamString,str) ):
        
        # Scenario 1 so just get value
        value = _ValueSingle( Mstar , Age , ParamString , ModelData=ModelData )
        
    else:
        
        # In this case, the output will be an array with up to three dimensions, so first make 3D array
        # and then remove dimensions with only one element
        
        # Get number of elements in all three directions
        try:
            nMstar = len(Mstar)
        except:
            nMstar = 1
        
        try:
            nAge = len(Age)
        except:
            nAge = 1
        
        if isinstance(ParamString,list):
            nParam = len(ParamString)
        else:
            nParam = 1
        
        # Make array
        value = np.zeros((nMstar,nAge,nParam))
        
        # Loop over values 
        for iMstar in range(0,nMstar):
            for iAge in range(0,nAge):
                for iParam in range(0,nParam):
                    
                    # Get values of Mstar, Age, and ParamString
                    if isinstance(Mstar,float):
                        MstarValue = Mstar
                    else:
                        MstarValue = Mstar[iMstar]
                        
                    if isinstance(Age,float):
                        AgeValue = Age
                    else:
                        AgeValue = Age[iAge]
                        
                    if isinstance(ParamString,str):
                        ParamStringValue = ParamString
                    else:
                        ParamStringValue = ParamString[iParam]
                    
                    # Now get the value
                    value[iMstar,iAge,iParam] = _ValueSingle( MstarValue , AgeValue , ParamStringValue , ModelData=ModelData )
            
        # Now get rid of unwanted dimensions
        value = np.squeeze(value)
        
    return value

def _ValueSingle(Mstar,Age,ParamString,ModelData=ModelDataDefault):
    """Takes stellar mass and age and parameter string, returns value of that parameter at that mass and age."""
    
    ## Make sure Mstar and age are floats
    #if not isinstance(Mstar,float):
        #misc._PrintErrorKill("argument Mstar must be float in call to Value")
    #if not isinstance(Age,float):
        #misc._PrintErrorKill("argument Age must be float in call to Value")
    
    # Make sure ParamString is indeed a string
    if not isinstance(ParamString,str):
        misc._PrintErrorKill("argument ParamString must be string in call to Value")
    
    # Make sure ParamString corresponds to a valid parameter
    if not ( ParamString in ModelData['ParamsAll'] ):
        misc._PrintErrorKill("parameter '"+ParamString+"' is not valid")
        
    # Check if an evolutionary track for this exact stellar mass is present
    if ( Mstar in ModelData ):
        
        # In this case, an evo track for this exact stellar mass is already loaded and all is needed is an interpolation
        # but first check the age is within the correct age limit
        _CheckAgeLimit( ModelData[Mstar]['Age'] , Age )
        value = _Interpolate1D( ModelData[Mstar]['Age'] , ModelData[Mstar][ParamString] , Age )
        
    else:
        
        # In this case, no evo track for this specific mass is loaded and so a 2D interpolation between values from two
        # separate mass bins is needed
        # Note, it is not assumed that the mass bins are loaded into the dictionary in ascending order
        
        # Make sure within mass limit
        _CheckMassLimit( ModelData['MstarAll'] , Mstar )
        
        # Get nearest mass bin below
        iMin = misc._getIndexLT( ModelData['MstarAll'] , Mstar )
        MstarMin = ModelData['MstarAll'][iMin]
        
        # Get nearest mass bin above
        iMax = misc._getIndexGT( ModelData['MstarAll'] , Mstar )
        MstarMax = ModelData['MstarAll'][iMax]
        
        # Check age ranges for both (max first since the upper limit is likely lower)
        _CheckAgeLimit( ModelData[MstarMax]['Age'] , Age )
        _CheckAgeLimit( ModelData[MstarMin]['Age'] , Age )
        
        # Do interpolation
        value = _Interpolate2D( MstarMin , MstarMax , Mstar ,
                                ModelData[MstarMin]['Age'] , ModelData[MstarMax]['Age'] , Age ,
                                ModelData[MstarMin][ParamString] , ModelData[MstarMax][ParamString] )
        
    return value

def _CheckAgeLimit(AgeArray,Age):
    """Takes age track and an age, outputs error and stops code if age is not within limits."""
    
    # Round the age to decimal places determined by nDecValue at top of file
    Age = round( Age , nDecValue )
    
    # Do check
    if not ( AgeArray[0] <= Age <= AgeArray[-1] ):
        misc._PrintErrorKill("input age "+str(Age)+" is not within limits of "+str(AgeArray[0])+" to "+str(AgeArray[-1]))
    
    return

def _CheckMassLimit(MstarArray,Mstar):
    """Takes array of masses and an age, outputs error and stops code if mass is not within limits."""
    
    if not ( np.min(MstarArray) <= Mstar <= np.max(MstarArray) ):
        misc._PrintErrorKill("input stellar mass "+str(Mstar)+" is not within limits of "+str(np.min(MstarArray))+" to "+str(np.max(MstarArray)))
    
    return

def _Interpolate2D(Z1,Z2,Z,Xarray1,Xarray2,X,Yarray1,Yarray2):
    """Takes two sets of 1D arrays for corresponding X and Y values, returns interpolated Y value corresponding to input X."""
    
    # Round the input values to decimal places determined by nDecValue at top of file
    Z = round( Z , nDecValue )
    
    # Do interpolations to get Y at X for both tracks
    Y1 = _Interpolate1D( Xarray1 , Yarray1 , X )
    Y2 = _Interpolate1D( Xarray2 , Yarray2 , X )
    
    # Do linear interpolation between Y1 and Y2 in Z direction
    mInterp = ( Y2 - Y1 ) / ( Z2 - Z1 )
    cInterp = Y1 - mInterp * Z1
    Y = mInterp * Z + cInterp
    
    return Y

def _Interpolate1D(Xarray,Yarray,X):
    """Takes 1D arrays for corresponding X and Y values, returns interpolated Y value corresponding to input X."""
    
    # Note that it is assumed here that Xarray is in ascending order and this won't work if it is not
    
    # Round the input values to decimal places determined by nDecValue at top of file
    X = round( X , nDecValue )  
    
    # Make sure X is in limits
    if not ( Xarray[0] <= X <= Xarray[-1] ):
        misc._PrintErrorKill("input value "+str(X)+" is not within limits of "+str(Xarray[0])+" to "+str(Xarray[-1]))
    
    # Get index of X closest to but smaller than value
    iMin = misc._getIndexLTordered(Xarray,X)
    iMax = iMin + 1
        
    # Check if X is an element in Xarray
    if ( Xarray[iMin] == X ):
        return Yarray[iMin]
    elif ( Xarray[iMax] == X ):
        return Yarray[iMax]
    else:
        
        # Do linear interpolation
        mInterp = ( Yarray[iMax] - Yarray[iMin] ) / ( Xarray[iMax] - Xarray[iMin] )
        cInterp = Yarray[iMin] - mInterp * Xarray[iMin]
        Y = mInterp * X + cInterp
    
    return Y

# The following functions are for individual parameters that can be called
def Rstar(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns stellar radius."""
    return Value( Mstar , Age , 'Rstar' , ModelData=ModelData )

def Lbol(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns bolometric luminosity."""
    return Value( Mstar , Age , 'Lbol' , ModelData=ModelData )

def Teff(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns effective temperature."""
    return Value( Mstar , Age , 'Teff' , ModelData=ModelData )

def Itotal(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns total moment of interia."""
    return Value( Mstar , Age , 'Itotal' , ModelData=ModelData )

def Icore(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns core moment of interia."""
    return Value( Mstar , Age , 'Icore' , ModelData=ModelData )

def Ienv(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns envelope moment of interia."""
    return Value( Mstar , Age , 'Ienv' , ModelData=ModelData )

def Mcore(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns core mass."""
    return Value( Mstar , Age , 'Mcore' , ModelData=ModelData )

def Menv(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns envelope mass."""
    return Value( Mstar , Age , 'Menv' , ModelData=ModelData )

def Rcore(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns core radius."""
    return Value( Mstar , Age , 'Rcore' , ModelData=ModelData )

def tauConv(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns convective turnover time."""
    return Value( Mstar , Age , 'tauConv' , ModelData=ModelData )

def dItotaldt(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns rate of change of total moment of inertia."""
    return Value( Mstar , Age , 'dItotaldt' , ModelData=ModelData )

def dIcoredt(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns rate of change of core moment of inertia."""
    return Value( Mstar , Age , 'dIcoredt' , ModelData=ModelData )

def dIenvdt(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns rate of change of envelope moment of inertia."""
    return Value( Mstar , Age , 'dIenvdt' , ModelData=ModelData )

def dMcoredt(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns rate of change of core mass."""
    return Value( Mstar , Age , 'dMcoredt' , ModelData=ModelData )

def dRcoredt(Mstar,Age,ModelData=ModelDataDefault):
    """Takes mass and age, returns rate of change of core radius."""
    return Value( Mstar , Age , 'dRcoredt' , ModelData=ModelData )
